import asyncio
from typing import Any
from unittest.mock import Mock, patch

import pytest

from pysmlight.ble_proxy import BLE_PROXY_VERSION, BleProxyClient, BleProxyProtocol
from pysmlight.const import BleProxyMode


class MockServerProtocol(asyncio.DatagramProtocol):
    def __init__(self) -> None:
        self.transport: asyncio.DatagramTransport | None = None
        self.received_packets: list[tuple[bytes, Any]] = []

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self.transport = transport  # type: ignore[assignment]

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        self.received_packets.append((data, addr))
        if len(data) >= 4 and data[1] == 0:
            use_port = int.from_bytes(data[2:4], byteorder="little")
            ack_pkt = b"\x00\x02"
            if self.transport:
                self.transport.sendto(ack_pkt, (addr[0], use_port))


@pytest.mark.asyncio
async def test_ble_proxy_protocol_ack() -> None:
    """Test that datagram_received processes an ACK action and triggers the on_ack callback."""
    callback = Mock()
    on_ack = Mock()
    protocol = BleProxyProtocol(callback, on_ack)
    protocol.datagram_received(b"\x00\x02", ("127.0.0.1", 12345))
    on_ack.assert_called_once()
    callback.assert_not_called()


@pytest.mark.asyncio
async def test_ble_proxy_protocol_invalid() -> None:
    """Test that datagram_received ignores invalid data."""
    callback = Mock()
    on_ack = Mock()
    protocol = BleProxyProtocol(callback, on_ack)
    protocol.datagram_received(b"\x01\x02", ("127.0.0.1", 12345))
    protocol.datagram_received(b"\x00", ("127.0.0.1", 12345))
    protocol.datagram_received(b"\x00\x03\x00\x00\x00\x00\x00", ("127.0.0.1", 12345))
    protocol.datagram_received(b"\x00\x05", ("127.0.0.1", 12345))
    mac_bytes = b"\x55\x44\x33\x22\x11\x00"
    incomplete_packet = (
        b"\x00\x03" + mac_bytes + b"\x01" + b"\xab" + b"\x0a" + b"\x02\x01\x06"
    )
    protocol.datagram_received(incomplete_packet, ("127.0.0.1", 12345))
    on_ack.assert_not_called()
    callback.assert_not_called()


@pytest.mark.asyncio
async def test_ble_proxy_protocol_valid_data() -> None:
    """Test that datagram_received successfully parses valid data."""
    callback = Mock()
    on_ack = Mock()
    protocol = BleProxyProtocol(callback, on_ack)
    mac_bytes = b"\x55\x44\x33\x22\x11\x00"
    addr_type = b"\x01"
    rssi = b"\xab"
    payload = b"\x02\x01\x06"
    packet = b"\x00\x03" + mac_bytes + addr_type + rssi + b"\x03" + payload
    protocol.datagram_received(packet, ("127.0.0.1", 12345))
    callback.assert_called_once_with(mac_bytes, -85, 1, payload)
    on_ack.assert_not_called()


@pytest.mark.asyncio
async def test_ble_proxy_protocol_bundled_data() -> None:
    """Test that datagram_received successfully parses bundled advertisements."""
    callback = Mock()
    on_ack = Mock()
    protocol = BleProxyProtocol(callback, on_ack)
    mac_bytes_1 = b"\x55\x44\x33\x22\x11\x00"
    payload_1 = b"\x02\x01\x06"
    packet_1 = b"\x00\x03" + mac_bytes_1 + b"\x01" + b"\xab" + b"\x03" + payload_1

    mac_bytes_2 = b"\xaa\xbb\xcc\xdd\xee\xff"
    payload_2 = b"\x05\x09\x4d\x79\x44\x65\x76"
    packet_2 = b"\x00\x03" + mac_bytes_2 + b"\x00" + b"\xb0" + b"\x07" + payload_2

    packet = packet_1 + packet_2
    protocol.datagram_received(packet, ("127.0.0.1", 12345))

    assert callback.call_count == 2
    callback.assert_any_call(mac_bytes_1, -85, 1, payload_1)
    callback.assert_any_call(mac_bytes_2, -80, 0, payload_2)
    on_ack.assert_not_called()


@pytest.mark.asyncio
async def test_ble_proxy_protocol_exception() -> None:
    """Test that exceptions inside datagram_received are caught and do not propagate."""
    callback = Mock(side_effect=ValueError("Callback error"))
    on_ack = Mock()
    protocol = BleProxyProtocol(callback, on_ack)
    mac_bytes = b"\x55\x44\x33\x22\x11\x00"
    addr_type = b"\x01"
    rssi = b"\xab"
    payload = b"\x02\x01\x06"
    packet = b"\x00\x03" + mac_bytes + addr_type + rssi + b"\x03" + payload
    protocol.datagram_received(packet, ("127.0.0.1", 12345))
    callback.assert_called_once()


@pytest.mark.asyncio
async def test_ble_proxy_client_success() -> None:
    """Test that BleProxyClient successfully starts, connects, pings, and stops cleanly."""
    loop = asyncio.get_running_loop()
    server_transport, server_protocol = await loop.create_datagram_endpoint(
        MockServerProtocol,
        local_addr=("127.0.0.1", 0),
    )
    server_port = server_transport.get_extra_info("sockname")[1]

    callback = Mock()
    client = BleProxyClient(
        esp32_ip="127.0.0.1",
        callback=callback,
        esp32_port=server_port,
    )

    original_sleep = asyncio.sleep

    async def mock_sleep(delay):
        if delay == 2.0:  # Backoff sleep or ping loop sleep
            await original_sleep(0.001)
        else:
            await original_sleep(delay)

    with patch("asyncio.sleep", side_effect=mock_sleep):
        await client.start()
        await original_sleep(0.001)
        assert client.transport is not None
        assert client.protocol is not None
        for _ in range(5):
            await original_sleep(0.01)
        client.stop()
        await asyncio.sleep(0.01)

    server_transport.close()
    assert len(server_protocol.received_packets) >= 2
    handshake_data, _ = server_protocol.received_packets[0]
    assert handshake_data[0] == 0
    assert handshake_data[1] == 0

    disconnect_found = False
    for data, _ in server_protocol.received_packets:
        if len(data) >= 2 and data[1] == 1:
            disconnect_found = True
            break
    assert disconnect_found


@pytest.mark.asyncio
async def test_ble_proxy_client_disconnect_exception() -> None:
    """Test that an exception during sending disconnect_packet in stop() is caught and ignored"""
    client = BleProxyClient(
        esp32_ip="127.0.0.1",
        callback=Mock(),
        esp32_port=5050,
    )
    mock_transport = Mock()
    mock_transport.sendto.side_effect = OSError("Transmit error")
    client.transport = mock_transport
    client.stop()
    mock_transport.close.assert_called_once()


@pytest.mark.asyncio
async def test_ble_proxy_client_timeout() -> None:
    """Test that BleProxyClient handles connection timeouts and executes proper exponential backoff retries."""
    callback = Mock()
    client = BleProxyClient(
        esp32_ip="127.0.0.1",
        callback=callback,
        esp32_port=1,
    )

    original_sleep = asyncio.sleep
    sleep_calls: list[float] = []

    async def mock_sleep(delay: float) -> None:
        sleep_calls.append(delay)
        await original_sleep(0.0001)

    async def mock_wait_for(fut: Any, timeout: float | None) -> Any:
        if asyncio.iscoroutine(fut):
            fut.close()
        raise TimeoutError()

    with (
        patch("asyncio.sleep", side_effect=mock_sleep),
        patch("asyncio.wait_for", side_effect=mock_wait_for),
        patch("pysmlight.ble_proxy._LOGGER") as mock_logger,
    ):
        await client.start()

        for _ in range(100):
            if len(sleep_calls) >= 2:
                break
            await original_sleep(0.001)

        client.stop()

    assert len(sleep_calls) >= 2
    assert sleep_calls[0] == 2.0
    assert sleep_calls[1] == 4.0
    mock_logger.warning.assert_called()


@pytest.mark.asyncio
async def test_ble_proxy_client_api_methods() -> None:
    """Test BleProxyClient's API methods set_scan_mode and set_active_window."""
    client = BleProxyClient(
        esp32_ip="127.0.0.1",
        callback=Mock(),
        esp32_port=5050,
    )
    mock_transport = Mock()
    client.transport = mock_transport

    client.set_scan_mode(BleProxyMode.BLE_PROXY_MODE_ACTIVE)
    mock_transport.sendto.assert_called_with(
        b"\x00\x04\x01",
        ("127.0.0.1", 5050),
    )

    client.set_scan_mode(BleProxyMode.BLE_PROXY_MODE_PASSIVE)
    mock_transport.sendto.assert_called_with(
        b"\x00\x04\x00",
        ("127.0.0.1", 5050),
    )

    client.set_active_window(timeout=120)
    mock_transport.sendto.assert_called_with(
        b"\x00\x05\x78\x00",
        ("127.0.0.1", 5050),
    )


@pytest.mark.asyncio
async def test_ble_proxy_client_api_methods_exceptions() -> None:
    """Test BleProxyClient's API methods raise SmlightConnectionError on socket failure."""
    from pysmlight.exceptions import SmlightConnectionError

    client = BleProxyClient(
        esp32_ip="127.0.0.1",
        callback=Mock(),
        esp32_port=5050,
    )
    mock_transport = Mock()
    mock_transport.sendto.side_effect = OSError("Socket error")
    client.transport = mock_transport

    with pytest.raises(SmlightConnectionError):
        client.set_scan_mode(BleProxyMode.BLE_PROXY_MODE_ACTIVE)

    with pytest.raises(SmlightConnectionError):
        client.set_active_window(timeout=120)


@pytest.mark.asyncio
async def test_ble_proxy_protocol_recovery() -> None:
    """Test that datagram_received successfully recovers from invalid prefix data or malformed packets."""
    callback = Mock()
    on_ack = Mock()
    protocol = BleProxyProtocol(callback, on_ack)

    packet_ack = b"\xff\xff" + bytes([BLE_PROXY_VERSION]) + b"\x02"
    protocol.datagram_received(packet_ack, ("127.0.0.1", 12345))
    on_ack.assert_called_once()
    on_ack.reset_mock()

    packet_invalid_action = (
        bytes([BLE_PROXY_VERSION]) + b"\x99" + bytes([BLE_PROXY_VERSION]) + b"\x02"
    )
    protocol.datagram_received(packet_invalid_action, ("127.0.0.1", 12345))
    on_ack.assert_called_once()
    on_ack.reset_mock()

    mac_bytes = b"\x55\x44\x33\x22\x11\x00"
    addr_type = b"\x01"
    rssi = b"\xab"
    payload = b"\x02\x01\x06"
    valid_data = (
        bytes([BLE_PROXY_VERSION])
        + b"\x03"
        + mac_bytes
        + addr_type
        + rssi
        + b"\x03"
        + payload
    )

    # First data packet has size specified as 100, but only 5 bytes of data are present
    truncated_packet = (
        bytes([BLE_PROXY_VERSION])
        + b"\x03"
        + mac_bytes
        + addr_type
        + rssi
        + b"\x64"
        + b"\x00\x00\x00\x00\x00"
    )
    protocol.datagram_received(truncated_packet + valid_data, ("127.0.0.1", 12345))
    callback.assert_called_once_with(mac_bytes, -85, 1, payload)
    callback.reset_mock()

    no_version_truncated = (
        bytes([BLE_PROXY_VERSION]) + b"\x03\xff\xff\xff\xff\xff\xff\x01\x9c\x64"
    )
    protocol.datagram_received(no_version_truncated, ("127.0.0.1", 12345))
    callback.assert_not_called()


@pytest.mark.asyncio
async def test_ble_proxy_client_ping_exception() -> None:
    """Test that BleProxyClient._send_ping catches OSError and logs warning."""
    client = BleProxyClient(
        esp32_ip="127.0.0.1",
        callback=Mock(),
        esp32_port=5050,
    )
    mock_transport = Mock()
    mock_transport.sendto.side_effect = OSError("Socket error")
    client.transport = mock_transport
    client.local_port = 12345

    from pysmlight.ble_proxy import _LOGGER as proxy_logger

    with patch.object(proxy_logger, "warning") as mock_warning:
        client._send_ping()
        mock_warning.assert_called_once()
        assert "Error sending ping" in mock_warning.call_args[0][0]



import asyncio
from collections.abc import Callable
import logging
import struct

from .const import BleProxyMode, ProxyAction
from .exceptions import SmlightConnectionError

_LOGGER = logging.getLogger(__name__)


# Structure of the fixed portion of the BLE proxy packet:
# - api_version (uint8)
# - api_action (uint8)
# - address (6 bytes)
# - address_type (uint8)
# - rssi (int8)
# - adv_data_len (uint8)
BLE_PROXY_HEADER_STRUCT = struct.Struct("<BB6sBbB")
BLE_PROXY_VERSION = 0


class BleProxyProtocol(asyncio.DatagramProtocol):
    """Protocol to handle incoming UDP packets from SLZB BLE Proxy server."""

    def __init__(
        self,
        callback: Callable[[bytes, int, int, bytes], None],
        on_ack: Callable[[], None],
    ) -> None:
        self.callback = callback
        self.on_ack = on_ack

    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        try:
            offset = 0
            while offset < len(data):
                if len(data) - offset < 2:
                    break

                version = data[offset]
                if version != BLE_PROXY_VERSION:
                    offset = data.find(BLE_PROXY_VERSION, offset + 1)
                    if offset == -1:
                        break
                    continue

                action = data[offset + 1]
                if action in (ProxyAction.ACK, ProxyAction.DATA):
                    if action == ProxyAction.ACK:
                        _LOGGER.debug("Received ACK packet from %s", addr)
                        self.on_ack()
                        offset += 2
                        continue

                    if action == ProxyAction.DATA:
                        if len(data) - offset < BLE_PROXY_HEADER_STRUCT.size:
                            break

                        (
                            _,
                            _,
                            mac_bytes,
                            address_type,
                            rssi,
                            adv_data_len,
                        ) = BLE_PROXY_HEADER_STRUCT.unpack_from(data, offset)

                        if (
                            len(data) - offset
                            < BLE_PROXY_HEADER_STRUCT.size + adv_data_len
                        ):
                            offset = data.find(BLE_PROXY_VERSION, offset + 1)
                            if offset == -1:
                                break
                            continue

                        raw_data = data[
                            offset + BLE_PROXY_HEADER_STRUCT.size : offset
                            + BLE_PROXY_HEADER_STRUCT.size
                            + adv_data_len
                        ]
                        if _LOGGER.isEnabledFor(logging.DEBUG):  # pragma: no cover
                            _LOGGER.debug(
                                "Parsed ADV packet from %s: mac_bytes=%s, rssi=%d, address_type=%d, adv_len=%d",
                                addr,
                                mac_bytes.hex(),
                                rssi,
                                address_type,
                                adv_data_len,
                            )
                        self.callback(mac_bytes, rssi, address_type, raw_data)
                        offset += BLE_PROXY_HEADER_STRUCT.size + adv_data_len
                        continue

                offset = data.find(BLE_PROXY_VERSION, offset + 1)
                if offset == -1:
                    break
        except Exception:
            _LOGGER.exception("Error parsing SLZB Bluetooth proxy packet from %s", addr)


class BleProxyClient:
    """Client to manage connection with SLZB BLE Proxy UDP server."""

    def __init__(
        self,
        esp32_ip: str,
        callback: Callable[[bytes, int, int, bytes], None],
        esp32_port: int = 5050,
    ) -> None:
        self.esp32_ip = esp32_ip
        self.esp32_port = esp32_port
        self.callback = callback
        self.transport: asyncio.DatagramTransport | None = None
        self.protocol: BleProxyProtocol | None = None
        self._connect_task: asyncio.Task | None = None
        self._ping_task: asyncio.Task | None = None
        self._connected_evt = asyncio.Event()
        self.running = False

    async def start(self) -> None:
        self.running = True
        self._connect_task = asyncio.create_task(self._connect_loop())

    async def _connect_loop(self) -> None:
        backoff = 2.0
        while self.running:
            try:
                loop = asyncio.get_running_loop()

                self.transport, self.protocol = await loop.create_datagram_endpoint(
                    lambda: BleProxyProtocol(self.callback, self._on_ack),
                    local_addr=("0.0.0.0", 0),
                )

                self.local_port = self.transport.get_extra_info("sockname")[1]
                self._send_ping()

                await asyncio.wait_for(self._connected_evt.wait(), timeout=2.0)
                self._ping_task = asyncio.create_task(self._ping_loop())
                break
            except (OSError, TimeoutError) as err:
                _LOGGER.warning(
                    "Connection to SLZB BLE Proxy failed, retrying in %.1fs: %s",
                    backoff,
                    err,
                )
                self.stop_transport()
                await asyncio.sleep(backoff)
                backoff = min(backoff * 2, 60.0)

    def _on_ack(self) -> None:
        self._connected_evt.set()

    def _send_ping(self) -> None:
        if self.transport and self.local_port != 0:
            ping_packet = struct.pack(
                "<BBH", BLE_PROXY_VERSION, ProxyAction.PING, self.local_port
            )
            try:
                self.transport.sendto(ping_packet, (self.esp32_ip, self.esp32_port))
            except OSError as ex:
                _LOGGER.warning("Error sending ping to SLZB BLE Proxy: %s", ex)

    async def _ping_loop(self) -> None:
        try:
            while True:
                await asyncio.sleep(2.0)
                self._send_ping()
        except asyncio.CancelledError:
            pass

    def stop_transport(self) -> None:
        if self.transport:
            try:
                disconnect_packet = struct.pack(
                    "<BB", BLE_PROXY_VERSION, ProxyAction.DISCONNECT
                )
                self.transport.sendto(
                    disconnect_packet, (self.esp32_ip, self.esp32_port)
                )
            except Exception:
                _LOGGER.exception("Error sending disconnect packet to SLZB BLE Proxy")
            finally:
                self.transport.close()
                self.transport = None
                self.protocol = None

    def stop(self) -> None:
        self.running = False
        if self._connect_task:
            self._connect_task.cancel()
            self._connect_task = None

        if self._ping_task:
            self._ping_task.cancel()
            self._ping_task = None

        self.stop_transport()
        self._connected_evt.clear()

    def set_scan_mode(self, mode: BleProxyMode) -> None:
        """Set scan mode."""
        if self.transport:
            packet = struct.pack(
                "<BBB", BLE_PROXY_VERSION, ProxyAction.SET_SCAN_MODE, mode
            )
            try:
                self.transport.sendto(packet, (self.esp32_ip, self.esp32_port))
            except OSError as ex:
                raise SmlightConnectionError(
                    f"Error sending scan mode to SLZB BLE Proxy: {ex}"
                ) from ex

    def set_active_window(self, timeout: int) -> None:
        """Request active scan window with the specified timeout in milliseconds (ms)."""
        if self.transport:
            packet = struct.pack(
                "<BBH", BLE_PROXY_VERSION, ProxyAction.REQ_ACTIVE_WINDOW, timeout
            )
            try:
                self.transport.sendto(packet, (self.esp32_ip, self.esp32_port))
            except OSError as ex:
                raise SmlightConnectionError(
                    f"Error requesting active scan window from SLZB BLE Proxy: {ex}"
                ) from ex



from __future__ import annotations

from dataclasses import dataclass, field
import re

from mashumaro import DataClassDictMixin, field_options

from .const import PERIPHERAL_MODELS, AmbiEffect, BleState, PppUSBState
from .payload import Payload


@dataclass
class Firmware(DataClassDictMixin):
    mode: str | None = None  # ESP|ZB|ESPs3
    type: int | None = None
    notes: str | None = None
    rev: str | None = None
    link: str | None = None
    ver: str | None = None
    dev: bool = False
    prod: bool = False
    baud: int | None = None

    def __post_init__(self):
        # zigbee firmware match key attributes
        if self.baud is not None:
            self.ver = self.rev
            self.dev = not self.prod

    def set_mode(self, mode):
        self.mode = mode


FirmwareList = list[Firmware] | None


@dataclass
class Radio(DataClassDictMixin):
    chip_index: int | None = None
    zb_channel: int | None = None
    zb_flash_size: int | None = None
    zb_hw: str | None = None
    zb_ram_size: int | None = None
    zb_version: str | None = None
    zb_type: int | None = None  # enum (coordinator, router, thread)
    radioModes: list[bool] | None = None


@dataclass
class BleFeatures(DataClassDictMixin):
    ble_enabled: bool | None = None
    proxy_enabled: bool | None = None


@dataclass
class Info(DataClassDictMixin):
    addons: dict[str, bool] = field(default_factory=dict)
    ble: BleFeatures | None = None
    coord_mode: int | None = None  # Enum
    device_ip: str | None = None
    fs_total: int | None = None
    fw_channel: str | None = None  # dev, beta or release
    hostname: str | None = None
    hw_version: str | None = None
    legacy_api: int = 0
    MAC: str | None = None
    model: str | None = None
    ram_total: int | None = None
    psram_total: int | None = None
    sw_version: str | None = None
    u_device: bool | None = None
    wifi_mode: int | None = None  # enum (off, client, AP etc)
    radios: list[Radio] = field(default_factory=list)
    zb_channel: int | None = None
    zb_flash_size: int | None = None
    zb_hw: str | None = None
    zb_ram_size: int | None = None
    zb_version: str | None = None
    zb_type: int | None = None  # enum (coordinator, router, thread)

    @classmethod
    def load_payload(cls, payload: Payload) -> Info:
        zb_version_val = int(payload.zb_version)
        return cls(
            device_ip=payload.device_ip,
            legacy_api=payload.legacy_api,
            MAC=payload.MAC,
            model=payload.model,
            fw_channel="release",
            sw_version=payload.sw_version,
            zb_hw=payload.zb_hw,
            zb_version=None if zb_version_val == -1 else str(zb_version_val),
        )

    @property
    def has_peripherals(self) -> bool:
        """Return true if the device supports peripheral features (e.g. Ambilight, buzzer, IR)."""
        return any(s in (self.model or "") for s in PERIPHERAL_MODELS)

    def check_zb_version(self, radio: Radio) -> Radio:
        if radio.zb_version is not None:
            if radio.zb_version in (-1, "-1"):
                radio.zb_channel = 2  # custom
                radio.zb_version = None
            else:
                radio.zb_version = str(radio.zb_version)
        return radio

    def __post_init__(self) -> None:
        if self.hw_version is not None:
            try:
                hw = int(self.hw_version)
                self.hw_version = f"{hw // 100}.{hw % 100:02d}"
            except ValueError:
                pass

        if self.model is not None:
            self.model = self.model.replace("P", "p")
            self.model = self.model.replace("MG", "Mg")

            if self.u_device is None:
                self.u_device = (
                    self.model.endswith("U") or "ultima" in self.model.lower()
                )

            # Zwave is optional module for Ultima, remove radio instance if not present
            if "Ultima" in self.model and not self.addons.get("zwave", False):
                self.radios = self.radios[:2]

        # Factory firmware may have invalid .plus suffix, convert to valid version
        if self.sw_version:
            if "plus" in self.sw_version:
                self.sw_version = re.sub(
                    r"\.plus(\d*)$",
                    lambda m: f".{m.group(1) if m.group(1) else '1'}",
                    self.sw_version,
                )
            elif self.sw_version.endswith(".dev"):
                self.sw_version = f"{self.sw_version}0"

        if not self.radios:
            # construct radio object for backward compatibility (v2.5.0)
            self.radios = [
                Radio(
                    chip_index=0,
                    zb_channel=self.zb_channel,
                    zb_flash_size=self.zb_flash_size,
                    zb_hw=self.zb_hw,
                    zb_ram_size=self.zb_ram_size,
                    zb_version=self.zb_version,
                    zb_type=self.zb_type,
                )
            ]

        for r in self.radios:
            self.check_zb_version(r)


@dataclass
class AmbilightPayload(DataClassDictMixin):
    ultLedMode: AmbiEffect | None = None
    ultLedColor: str | None = None
    ultLedColor2: str | None = None
    ultLedSpeed: int | None = None
    ultLedBri: int | None = None
    ultLedDir: int | None = None


@dataclass
class BleSession(DataClassDictMixin):
    state: BleState | None = None
    proxy_connected: bool | None = None


@dataclass
class Sensors(DataClassDictMixin):
    esp32_temp: float | None = None
    zb_temp: float | None = None
    zb_temp2: float | None = None
    uptime: int = 0
    socket_uptime: int | None = None
    socket_uptime2: int | None = field(
        default=None, metadata=field_options(alias="socket2_uptime")
    )
    socket_uptime3: int | None = field(
        default=None, metadata=field_options(alias="socket3_uptime")
    )
    otbr_uptime: int | None = None
    ram_usage: int | None = None
    psram_usage: int | None = None
    fs_used: int | None = None
    ethernet: bool = False
    wifi_connected: bool = False
    wifi_status: int | str | None = None  # enum (off, client, AP etc)
    vpn_status: bool = False

    # toggle states:
    disable_leds: bool | None = None
    night_mode: bool | None = None
    auto_zigbee: bool | None = None
    vpn_enabled: bool | None = None

    ble: BleSession | None = None

    # Ultima additional sensors:
    ambilight: AmbilightPayload | None = None
    lte_detect: bool | None = None
    lte_state: PppUSBState | None = None

    def __post_init__(self) -> None:
        for field_name in (
            "socket_uptime",
            "socket_uptime2",
            "socket_uptime3",
            "otbr_uptime",
        ):
            value = getattr(self, field_name)
            if value == 0:
                setattr(self, field_name, None)


@dataclass
class SettingsEvent(DataClassDictMixin):
    page: int | None = None
    origin: str | None = None
    needReboot: bool = False
    setting: dict[str, bool | int] | None = None


@dataclass
class IRPayload(DataClassDictMixin):
    code: str | None = None
    freq: int | None = None

    @classmethod
    def from_raw_timings(cls, timings: list[int], freq: int | None = None) -> IRPayload:
        """Create an IRPayload from raw pulse timings in microseconds."""
        timings_hex: list[str] = []
        for interval in timings:
            ticks = int(interval / 50)
            while ticks > 255:
                timings_hex.extend(["ff", "00"])
                ticks -= 255
            timings_hex.append(f"{max(1, ticks):02x}")
        freq_khz = round(freq / 1000) if freq else 38
        return cls(code="".join(timings_hex), freq=freq_khz)

    def to_raw_timings(self) -> list[int]:
        """Convert IRPayload code to raw pulse timings in microseconds."""
        if not self.code:
            raise ValueError("IRPayload code is empty")

        timings: list[int] = []
        current_ticks = 0
        skip_next = False

        hex_pairs = [self.code[i : i + 2] for i in range(0, len(self.code), 2)]

        for i, pair in enumerate(hex_pairs):
            if skip_next:
                skip_next = False
                continue

            ticks = int(pair, 16)

            if ticks == 255 and i + 1 < len(hex_pairs) and hex_pairs[i + 1] == "00":
                current_ticks += 255
                skip_next = True
            else:
                timings.append((current_ticks + ticks) * 50)
                current_ticks = 0

        return timings


@dataclass
class BuzzerPayload(DataClassDictMixin):
    code: str | None = None



"""Tests for session initialization."""

from aiohttp import ClientSession

from pysmlight.web import Api2

host = "slzb-06.local"


async def test_init_with_session() -> None:
    """Test Api2 initialization with provided session."""
    async with ClientSession() as session:
        client = Api2(host, session=session)
        assert client.session is session
        assert client.close_session is False
        assert client.sse is not None


async def test_init_without_session() -> None:
    """Test Api2 initialization without provided session."""
    client = Api2(host)
    try:
        assert client.session is not None
        assert client.close_session is True
        assert client.sse is not None
    finally:
        await client.close()


async def test_init_with_sse() -> None:
    """Test Api2 initialization with provided sse client."""
    from pysmlight.sse import sseClient

    async with ClientSession() as session:
        sse = sseClient(host, session)
        client = Api2(host, session=session, sse=sse)
        assert client.session is session
        assert client.sse is sse



"""Test sending actions to SLZB devices."""

from __future__ import annotations

import logging
from urllib.parse import parse_qs

from aiohttp import ClientSession
from aresponses import ResponsesMockServer

from pysmlight import Api2
from pysmlight.const import Actions, AmbiEffect, Pages
from pysmlight.models import AmbilightPayload, BuzzerPayload, IRPayload

_LOGGER = logging.getLogger(__name__)

host = "slzb-06.local"


def aresponses_fixture(aresponses: ResponsesMockServer):
    aresponses.add(
        host,
        "/settings/saveParams",
        "POST",
        aresponses.Response(
            status=200, headers={"Content-Type": "application/json"}, text="ok"
        ),
    )
    aresponses.add(
        host,
        "/api2",
        "GET",
        aresponses.Response(
            status=200, headers={"Content-Type": "application/json"}, text="ok"
        ),
    )


async def test_actions_ambilight(aresponses: ResponsesMockServer) -> None:
    """Test sending Ambilight command to SLZB devices."""

    aresponses_fixture(aresponses)

    async with ClientSession() as session:
        client = Api2(host, session=session)

        payload = AmbilightPayload(
            ultLedMode=AmbiEffect.WSULT_SOLID,
            ultLedColor="#ff0000",
            ultLedColor2="#00ff00",
            ultLedSpeed=50,
            ultLedBri=255,
            ultLedDir=1,
        )

        await client.actions.ambilight(payload)

        req = aresponses.history[0].request
        assert req.method == "POST"
        assert req.path == "/settings/saveParams"

        body = await req.text()
        data = parse_qs(body)

        assert data["pageId"][0] == str(Pages.API2_PAGE_AMBILIGHT.value)
        assert data["ultLedMode"][0] == str(AmbiEffect.WSULT_SOLID.value)
        assert data["ultLedColor"][0] == "#ff0000"
        assert data["ultLedColor2"][0] == "#00ff00"
        assert data["ultLedSpeed"][0] == "50"
        assert data["ultLedBri"][0] == "255"
        assert data["ultLedDir"][0] == "1"


async def test_actions_buzzer(aresponses: ResponsesMockServer) -> None:
    """Test sending Buzzer command to SLZB devices."""

    async def response_handler(request):
        body = await request.text()
        data = parse_qs(body)
        assert data["action"][0] == str(Actions.API_BUZZER.value)
        assert data["code"][0] == "Arkanoid:d=4,o=5,b=140:8g6,16p"
        return aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text="ok",
        )

    aresponses.add(host, "/api2", "POST", response_handler)

    async with ClientSession() as session:
        client = Api2(host, session=session)

        payload = BuzzerPayload(code="Arkanoid:d=4,o=5,b=140:8g6,16p")

        res = await client.actions.buzzer(payload)

        req = aresponses.history[0].request
        assert req.method == "POST"

        assert res is True


async def test_actions_send_ir_code(aresponses: ResponsesMockServer) -> None:
    """Test sending IR command to SLZB devices."""

    aresponses_fixture(aresponses)

    async with ClientSession() as session:
        client = Api2(host, session=session)

        payload = IRPayload(code="C2B0A9")

        await client.actions.send_ir_code(payload)

        req = aresponses.history[0].request
        assert req.method == "POST"

        body = await req.text()
        data = parse_qs(body)

        assert data["action"][0] == str(Actions.API_IR.value)
        assert data["code"][0] == "C2B0A9"


async def test_actions_get_ir_code(aresponses: ResponsesMockServer) -> None:
    """Test getting IR code from SLZB devices."""

    async def response_handler(request):
        data = request.query
        assert data["pageId"] == str(Pages.API2_PAGE_IR.value)
        return aresponses.Response(
            status=200,
            headers={"Content-Type": "text/plain"},
            text="C2B0A9",
        )

    aresponses.add(host, "/api2", "GET", response_handler)

    async with ClientSession() as session:
        client = Api2(host, session=session)

        payload = IRPayload()
        res = await client.actions.get_ir_code(payload)

        req = aresponses.history[0].request
        assert req.method == "GET"

        assert res == "C2B0A9"



"""Tests for retrieving device information from SLZB-06x devices."""

import json
from unittest.mock import patch

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientConnectionError
from aresponses import ResponsesMockServer
import pytest
from syrupy.assertion import SnapshotAssertion

from pysmlight import Api2, Info, Sensors
from pysmlight.const import Settings
from pysmlight.exceptions import SmlightAuthError, SmlightConnectionError

from . import load_fixture

host = "slzb-06.local"


async def test_info_device_info(
    aresponses: ResponsesMockServer, snapshot: SnapshotAssertion
) -> None:
    """Test getting SLZB device information."""
    aresponses.add(
        host,
        "/ha_info",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("slzb-06-info.json"),
        ),
    )
    async with ClientSession() as session:
        client = Api2(host, session=session)
        info: Info = await client.get_info()
        assert info == snapshot


async def test_info_device_mr_info(
    aresponses: ResponsesMockServer, snapshot: SnapshotAssertion
) -> None:
    """Test getting SLZB device information for multi radios."""
    aresponses.add(
        host,
        "/ha_info",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("slzb-06-info-radios.json"),
        ),
    )
    async with ClientSession() as session:
        client = Api2(host, session=session)
        info: Info = await client.get_info()
        assert info == snapshot


async def test_ultima_info(
    aresponses: ResponsesMockServer, snapshot: SnapshotAssertion
) -> None:
    """Test getting SLZB Ultima device information."""
    aresponses.add(
        host,
        "/ha_info",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("slzb-ultima-info.json"),
        ),
    )
    async with ClientSession() as session:
        client = Api2(host, session=session)
        info: Info = await client.get_info()
        assert info == snapshot


async def test_info_get_auth_fail(aresponses: ResponsesMockServer) -> None:
    """Test getting SLZB device information."""
    aresponses.add(
        host,
        "/ha_info",
        "GET",
        aresponses.Response(
            status=401,
            headers={"Content-Type": "application/json"},
            text="wrong login or password",
        ),
    )
    async with ClientSession() as session:
        client = Api2(host, session=session)
        with pytest.raises(SmlightAuthError):
            await client.get_info()


@pytest.mark.asyncio
@patch("aiohttp.ClientSession.get")
async def test_info_get_connector_error(mock_get) -> None:
    """Test getting SLZB device information."""

    async def mock_raise(*args, **kwargs):
        raise ClientConnectionError("Mocked connection error")

    mock_get.return_value.__aenter__.side_effect = mock_raise
    async with ClientSession() as session:
        client = Api2(host, session=session)

        with pytest.raises(SmlightConnectionError):
            await client.get_info()
        with pytest.raises(SmlightConnectionError):
            await client.get_sensors()


@pytest.mark.asyncio
@patch("aiohttp.ClientSession.post")
async def test_info_post_connector_error(mock_post) -> None:
    """Test getting SLZB device information."""

    async def mock_raise(*args, **kwargs):
        raise ClientConnectionError("Mocked connection error")

    mock_post.return_value.__aenter__.side_effect = mock_raise
    async with ClientSession() as session:
        client = Api2(host, session=session)
        _page, _toggle = Settings.NIGHT_MODE.value
        with pytest.raises(SmlightConnectionError):
            await client.set_toggle(_page, _toggle, True)


async def test_info_sensors(
    aresponses: ResponsesMockServer, snapshot: SnapshotAssertion
) -> None:
    """Test getting SLZB sensor data."""
    aresponses.add(
        host,
        "/ha_sensors",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("slzb-06-sensors.json"),
        ),
    )
    async with ClientSession() as session:
        client = Api2(host, session=session)
        sensors = await client.get_sensors()
        assert sensors == snapshot


async def test_ultima_sensors(
    aresponses: ResponsesMockServer, snapshot: SnapshotAssertion
) -> None:
    """Test getting SLZB Ultima sensor data."""
    aresponses.add(
        host,
        "/ha_sensors",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("slzb-ultima-sensors.json"),
        ),
    )
    async with ClientSession() as session:
        client = Api2(host, session=session)
        sensors = await client.get_sensors()
        assert sensors == snapshot
        assert sensors.auto_zigbee is False


@pytest.mark.parametrize(
    "field",
    ["socket_uptime", "socket_uptime2", "socket_uptime3", "otbr_uptime"],
)
def test_sensors_radio_uptime_zero_normalized_to_none(field: str) -> None:
    """Radio uptime value 0 is normalized to None."""
    sensors = Sensors(**{field: 0})
    assert getattr(sensors, field) is None


def test_sensors_radio_uptime_preserves_positive() -> None:
    """Positive radio uptime values are kept as-is."""
    sensors = Sensors(
        socket_uptime=10,
        socket_uptime2=20,
        socket_uptime3=30,
        otbr_uptime=40,
    )
    assert sensors.socket_uptime == 10
    assert sensors.socket_uptime2 == 20
    assert sensors.socket_uptime3 == 30
    assert sensors.otbr_uptime == 40


def test_sensors_socket_uptime_device_keys() -> None:
    """Device payload keys map onto the normalized property names."""
    sensors = Sensors.from_dict({"socket2_uptime": 123, "socket3_uptime": 45})
    assert sensors.socket_uptime2 == 123
    assert sensors.socket_uptime3 == 45


async def test_info_legacy_info(aresponses: ResponsesMockServer) -> None:
    """Test getting legacy device info from devices with old firmware."""
    headers = {
        "Content-Type": "application/json",
        "respValuesArr": json.dumps(json.loads(load_fixture("slzb-06-resparr.json"))),
    }
    aresponses.add(
        host,
        "/ha_info",
        "GET",
        aresponses.Response(
            status=404,
            headers={"Content-Type": "application/json"},
        ),
    )
    aresponses.add(
        host,
        "/api2",
        "GET",
        aresponses.Response(status=200, headers=headers, text="Some html"),
    )
    async with ClientSession() as session:
        client = Api2(host, session=session)
        info: Info = await client.get_info()
        assert info

        assert info.sw_version == "v2.0.20"
        assert info.zb_version is None
        assert info.legacy_api == 1


async def test_info_legacy_info2(aresponses: ResponsesMockServer) -> None:
    """Test getting legacy device info from devices with old firmware."""
    headers = {
        "Content-Type": "application/json",
        "respValuesArr": json.dumps(
            json.loads(load_fixture("slzb-06-resparr-0.9.9.json"))
        ),
    }
    aresponses.add(
        host,
        "/ha_info",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text="URL NOT FOUND",
        ),
    )
    aresponses.add(
        host,
        "/api",
        "GET",
        aresponses.Response(status=200, headers=headers, text="Some html"),
    )
    async with ClientSession() as session:
        client = Api2(host, session=session)
        info: Info = await client.get_info()
        assert info

        assert info.sw_version == "0.9.9"
        assert info.legacy_api == 2


async def test_info_invalid_versions() -> None:
    """Test conversion of invalid firmware versions (per awesomeVersion)

    factory firmware that may have .plus suffix
    development firmware that may have .dev suffix."""
    info = Info(sw_version="v2.3.6.plus")
    assert info.sw_version == "v2.3.6.1"
    info = Info(sw_version="v2.3.6.plus2")
    assert info.sw_version == "v2.3.6.2"
    info = Info(sw_version="v2.8.0.dev")
    assert info.sw_version == "v2.8.0.dev0"


async def test_info_ultima_zwave() -> None:
    """Test Ultima only has 2 radios without zwave addon installed."""
    info = Info.from_dict(
        {
            "model": "SLZB-Ultima4",
            "addons": {"zwave": False},
            "radios": [
                {"zb_hw": "Radio1"},
                {"zb_hw": "Radio2"},
                {"zb_hw": "Radio3 (Z-Wave)"},
            ],
        }
    )

    assert len(info.radios) == 2
    assert info.radios[0].zb_hw == "Radio1"
    assert info.radios[1].zb_hw == "Radio2"


@pytest.mark.parametrize(
    ("model", "expected"),
    [
        ("SLZB-Ultima4", True),
        ("SLZB-Ultima3", True),
        ("SLZB-06p10", False),
        (None, False),
    ],
)
def test_info_has_peripherals(model: str | None, expected: bool) -> None:
    """Test has_peripherals returns True only for peripheral-capable models."""
    info = Info(model=model)
    assert info.has_peripherals is expected


@pytest.mark.parametrize(
    ("hw_version", "expected"),
    [
        (104, "1.04"),
        ("1.04", "1.04"),
        (None, None),
    ],
)
def test_info_hw_version_formatting(
    hw_version: int | str | None, expected: str | None
) -> None:
    """Test hw_version is formatted correctly from raw integer or pre-formatted string."""
    info = Info(hw_version=hw_version)
    assert info.hw_version == expected



__all__ = [
    "Api2",
    "CmdWrapper",
    "Firmware",
    "Info",
    "Radio",
    "Sensors",
    "SettingsEvent",
    "BleProxyClient",
    "BleProxyMode",
    "BleProxyProtocol",
]

from pysmlight.ble_proxy import BleProxyClient, BleProxyProtocol
from pysmlight.const import BleProxyMode
from pysmlight.models import Radio, SettingsEvent
from pysmlight.web import Api2, CmdWrapper, Firmware, Info, Sensors

try:
    from pysmlight._version import __version__
except ImportError:  # pragma: no cover
    __version__ = "0.0.0.dev0"



import pytest

from pysmlight.models import IRPayload


@pytest.mark.parametrize(
    "timings, freq, expected_code, expected_freq",
    [
        ([500, 1000], 38000, "0a14", 38),  # Regular conversion
        ([500], None, "0a", 38),  # Missing optional frequency
        (
            [26000],
            38000,
            "ff00ff000a",
            38,
        ),  # Large ticks greater than 255
    ],
)
def test_ir_payload_from_raw_timings(timings, freq, expected_code, expected_freq):
    payload = IRPayload.from_raw_timings(timings, freq=freq)
    assert payload.code == expected_code
    assert payload.freq == expected_freq


def test_ir_payload_to_raw_timings_empty():
    payload = IRPayload()
    with pytest.raises(ValueError, match="IRPayload code is empty"):
        payload.to_raw_timings()


@pytest.mark.parametrize(
    "code, expected_timings",
    [
        ("0a14", [500, 1000]),  # Basic usage
        ("ff00ff000a", [26000]),  # Large ticks greater than 255
        ("aaff", [8500, 12750]),  # Edge case ending perfectly with 255
    ],
)
def test_ir_payload_to_raw_timings(code, expected_timings):
    payload = IRPayload(code=code)
    assert payload.to_raw_timings() == expected_timings
